"""Container for caching the reports data used in views."""

import datetime as dt
from typing import Optional

from django.core.cache import cache
from django.db.models import Count, F, Max, Min, Q, Sum, Value
from django.db.models.functions import Concat, TruncMinute
from django.urls import reverse
from django.utils import timezone

from ..app_settings import (
    TASKMONITOR_HOUSEKEEPING_FREQUENCY,
    TASKMONITOR_REPORTS_MAX_AGE,
    TASKMONITOR_REPORTS_MAX_TOP,
)
from ..models import TaskLog

CACHE_KEY = "TASKMONITOR_REPORTS_DATA"


def data() -> dict:
    """Return the cached reports data."""
    context = cache.get_or_set(CACHE_KEY, _calc_data, timeout=_timeout())
    ttl = cache.ttl(CACHE_KEY)
    context["last_update_at"] = _last_update_at(ttl)
    context["next_update_at"] = _next_update_at(ttl)
    return context


def _timeout() -> int:
    """Timeout in seconds."""
    return TASKMONITOR_REPORTS_MAX_AGE * 60


def _last_update_at(ttl) -> Optional[dt.datetime]:
    """When the cache was last updated or None if there is no cache."""
    if not ttl:
        return None
    return timezone.now() - dt.timedelta(seconds=max(0, _timeout() - ttl))


def _next_update_at(ttl) -> Optional[dt.datetime]:
    """When the cache will be updated next (earliest) or None if no cache."""
    if not ttl:
        return None
    duration = TASKMONITOR_HOUSEKEEPING_FREQUENCY * 60 / 2 + ttl
    return timezone.now() + dt.timedelta(seconds=duration)


def refresh_cache() -> None:
    """Refresh the cache."""
    cache.set(CACHE_KEY, _calc_data(), timeout=_timeout())


def clear_cache() -> None:
    """Clear the cache."""
    cache.delete(CACHE_KEY)


def _calc_data() -> dict:
    """Calculate the report data."""
    now = timezone.now()
    oldest_date = TaskLog.objects.aggregate(oldest=Min("timestamp"))["oldest"]
    youngest_date = TaskLog.objects.aggregate(youngest=Max("timestamp"))["youngest"]
    total_runtime, total_runtime_date = _calc_total_runtime(now)
    total_runs = TaskLog.objects.count()
    changelist_url = reverse("admin:taskmonitor_tasklog_changelist")
    context = {
        "oldest_date": oldest_date,
        "youngest_date": youngest_date,
        "total_runs": total_runs,
        "total_runtime_date": total_runtime_date,
        "task_totals_by_state": _calc_task_totals_by_state(total_runs, changelist_url),
        "task_runs_per_app": _calc_task_runs_per_app(total_runs, changelist_url),
        "tasks_top_runs": _calc_tasks_top_runs(total_runs, changelist_url),
        "tasks_top_runtime": _calc_tasks_top_runtime(total_runtime, changelist_url),
        "tasks_top_failed": _calc_tasks_top_failed(changelist_url),
        "tasks_top_retried": _calc_tasks_top_retried(changelist_url),
        "tasks_throughput": _calc_tasks_throughput(now),
        "tasks_throughput_2": _calc_tasks_throughput_2(now),
        "MAX_TOP": TASKMONITOR_REPORTS_MAX_TOP,
    }
    return context


def _calc_total_runtime(now):
    total_runtime = TaskLog.objects.aggregate(total_runtime=Sum("runtime"))[
        "total_runtime"
    ]
    try:
        total_runtime_date = now - dt.timedelta(seconds=total_runtime)
    except TypeError:
        total_runtime_date = None
    return total_runtime, total_runtime_date


def _calc_task_totals_by_state(total_runs, changelist_url):
    if not total_runs:
        return None
    return [
        {
            "name": state.label,
            "y": TaskLog.objects.filter(state=state.value).count(),
            "url": f"{changelist_url}?state__exact={state}",
        }
        for state in TaskLog.State
    ]


def _calc_task_runs_per_app(total_runs, changelist_url):
    if not total_runs:
        return None
    return list(
        TaskLog.objects.values(name=F("app_name"))
        .annotate(y=Count("pk"))
        .annotate(url=Concat(Value(f"{changelist_url}?app_name="), F("name")))
        .order_by("-y")
    )


def _calc_tasks_top_runs(total_runs, changelist_url):
    if not total_runs:
        return None
    return list(
        TaskLog.objects.values(name=F("task_name"))
        .annotate(y=Count("pk"))
        .annotate(url=Concat(Value(f"{changelist_url}?task_name="), F("name")))
        .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
    )


def _calc_tasks_top_runtime(total_runtime, changelist_url):
    if not total_runtime:
        return None
    return list(
        TaskLog.objects.values(name=F("task_name"))
        .annotate(y=Max("runtime"))
        .annotate(url=Concat(Value(f"{changelist_url}?o=5&task_name="), F("name")))
        .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
    )


def _calc_tasks_top_failed(changelist_url):
    total_failed = TaskLog.objects.filter(state=TaskLog.State.FAILURE).count()
    if not total_failed:
        return None
    return list(
        TaskLog.objects.filter(state=TaskLog.State.FAILURE)
        .values(name=F("task_name"))
        .annotate(y=Count("pk"))
        .annotate(
            url=Concat(Value(f"{changelist_url}?state__exact=3&task_name="), F("name"))
        )
        .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
    )


def _calc_tasks_top_retried(changelist_url):
    total_retried = TaskLog.objects.filter(state=TaskLog.State.RETRY).count()
    if not total_retried:
        return None
    return list(
        TaskLog.objects.filter(state=TaskLog.State.RETRY)
        .values(name=F("task_name"))
        .annotate(y=Count("pk"))
        .annotate(
            url=Concat(Value(f"{changelist_url}?state__exact=2&task_name="), F("name"))
        )
        .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
    )


def _calc_tasks_throughput(now):
    tasklogs_not_failed = TaskLog.objects.exclude(state=TaskLog.State.FAILURE)
    tasks_throughput = []
    average_last_hours = dict()
    for hours in [1, 3, 6, 12, 24]:
        average_last_hours[hours] = tasklogs_not_failed.filter(
            timestamp__gt=now - dt.timedelta(hours=hours)
        ).avg_throughput()
    for hours, amount in average_last_hours.items():
        tasks_throughput.append(
            {"name": f"Average last {hours} hours", "amount": amount}
        )
    average_overall = tasklogs_not_failed.avg_throughput()
    tasks_throughput.append({"name": "Average overall", "amount": average_overall})
    peak_overall = tasklogs_not_failed.max_throughput()
    tasks_throughput.append({"name": "Peak overall", "amount": peak_overall})
    return tasks_throughput


def _calc_tasks_throughput_2(now):
    result = (
        TaskLog.objects.annotate(x=TruncMinute("timestamp"))
        .values("x")
        .annotate(succeeded=Count("id", filter=Q(state=TaskLog.State.SUCCESS)))
        .annotate(retried=Count("id", filter=Q(state=TaskLog.State.RETRY)))
        .annotate(failed=Count("id", filter=Q(state=TaskLog.State.FAILURE)))
    )
    return list(result)



import concurrent.futures
import functools
import itertools
import json

import redis

from django.conf import settings

PRIORITY_SEP = "\x06\x16"
DEFAULT_PRIORITY_STEPS = range(10)


def _redis_client():
    """Fetch the Redis client for the celery broker."""
    return redis.from_url(settings.BROKER_URL)


def _queue_base_name() -> str:
    """Base name of celery queue."""
    return getattr(settings, "CELERY_DEFAULT_QUEUE", "celery")


def _queue_names(base_name: str = None) -> list:
    """List of all queue names incl. the dedicated queue names for each priority."""
    if not base_name:
        base_name = _queue_base_name()
    names = [
        f"{base_name}{PRIORITY_SEP}{priority}" for priority in DEFAULT_PRIORITY_STEPS
    ]
    names = [base_name] + names
    return names


def queue_length() -> list:
    """Length of the celery queue."""
    r = _redis_client()
    return sum(r.llen(queue_name) for queue_name in _queue_names())


def _fetch_tasks_from_queue(r: redis.Redis, queue_name: str) -> list:
    """Fetch tasks from given queue and return ordered
    with oldest task in first position.
    """
    tasks_raw = r.lrange(queue_name, 0, -1)
    tasks = [json.loads(obj.decode("utf8")) for obj in tasks_raw]
    return reversed(tasks)


def fetch_tasks() -> list:
    """Fetch all tasks from queues and return as combined list."""
    _fetch_func = functools.partial(_fetch_tasks_from_queue, _redis_client())
    queue_names = _queue_names()
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=len(queue_names)
    ) as executor:
        tasks_raw = executor.map(_fetch_func, queue_names)
    return list(itertools.chain(*tasks_raw))



import datetime as dt
import traceback as tb
from typing import List
from uuid import UUID

from django.db import models
from django.db.models import Avg, Count, Max
from django.db.models.functions import TruncMinute
from django.utils import timezone

from .core import celery_queues
from .helpers import extract_app_name


class QuerySetQueryStub:
    def __init__(self) -> None:
        self.select_related = None
        self.order_by = []


class ListAsQuerySet(list):
    def __init__(self, *args, model, distinct=False, **kwargs):
        self.model = model
        self.query = QuerySetQueryStub()
        self.distinct_enabled = distinct
        super().__init__(*args, **kwargs)
        self._id_mapper = {str(obj.id): n for n, obj in enumerate(self)}

    def get(self, *args, **kwargs):
        if "id" in kwargs:
            try:
                return self[self._id_mapper[str(kwargs["id"])]]
            except KeyError:
                raise self.model.DoesNotExist from None
        raise self.model.DoesNotExist

    def distinct(self):
        return ListAsQuerySet(list(set(self)), model=self.model, distinct=True)

    def values(self, *args):
        result = [
            {k: v for k, v in obj.__dict__.items() if not args or k in args}
            for obj in self
        ]
        return result

    def values_list(self, *args, **kwargs):
        items = [tuple(obj.values()) for obj in self.values(*args)]
        if kwargs.get("flat"):
            items = [obj[0] for obj in items]
            if self.distinct_enabled:
                return list(dict.fromkeys(items))
            return items
        return items

    def first(self):
        return self[0] if self else None

    def filter(self, *args, **kwargs):
        if kwargs:
            new_list = []
            for obj in self:
                if all([getattr(obj, key) == value for key, value in kwargs.items()]):
                    new_list.append(obj)
            return ListAsQuerySet(new_list, model=self.model)
        return self

    def order_by(self, *args, **kwargs):
        if args:
            for prop in reversed(args):
                if prop[0:1] == "-":
                    reverse = True
                    prop = prop[1:]
                else:
                    reverse = False
                self.sort(key=lambda d: getattr(d, prop), reverse=reverse)
        return self

    def count(self):
        return len(self)

    def _clone(self):
        return self


class QueuedTaskQuerySet(models.QuerySet):
    def count(self):
        return celery_queues.queue_length()


class QueuedTaskManagerBase(models.Manager):
    def get_queryset(self):
        from .models import QueuedTask

        objs = []
        for position, obj in enumerate(celery_queues.fetch_tasks()):
            try:
                objs.append(QueuedTask.create_from_dict(obj, position))
            except ValueError:
                pass
        return ListAsQuerySet(objs, model=QueuedTask)


QueuedTaskManager = QueuedTaskManagerBase.from_queryset(QueuedTaskQuerySet)


class TaskLogQuerySet(models.QuerySet):
    def csv_line_generator(self, fields: List[str]):
        """Return the tasklogs for a CSV file line by line.
        And return the field names as first line.
        """
        yield [field.name for field in fields]
        for obj in self.iterator():
            values = []
            for field in fields:
                if field.choices:
                    value = getattr(obj, f"get_{field.name}_display")()
                else:
                    value = getattr(obj, field.name)
                # if callable(value):
                #     try:
                #         value = value() or ""
                #     except Exception:
                #         value = "Error retrieving value"
                if value is None:
                    value = ""
                values.append(value)
            yield values

    def aggregate_timestamp_trunc(self):
        """Aggregate timestamp trunc."""
        return (
            self.annotate(timestamp_trunc=TruncMinute("timestamp"))
            .values("timestamp_trunc")
            .annotate(task_runs=Count("id"))
        )

    def max_throughput(self) -> int:
        """Calculate the maximum throughput in task executions per minute."""
        qs = self.aggregate_timestamp_trunc().aggregate(Max("task_runs"))
        return qs["task_runs__max"]

    def avg_throughput(self) -> float:
        """Calculate the average throughput in task executions per minute."""
        qs = self.aggregate_timestamp_trunc().aggregate(Avg("task_runs"))
        return qs["task_runs__avg"]


class TaskLogManagerBase(models.Manager):
    def create_from_task(
        self,
        *,
        task_id: str,
        task_name: str,
        state: int,
        retries: int,
        priority: int,
        received: dt.datetime = None,
        started: dt.datetime = None,
        parent_id: str = None,
        exception=None,
    ) -> models.Model:
        """Create new object from a celery task."""
        args = {
            "app_name": extract_app_name(task_name),
            "priority": priority,
            "parent_id": UUID(parent_id) if parent_id else None,
            "received": received,
            "retries": retries,
            "started": started,
            "state": state,
            "task_id": UUID(task_id),
            "task_name": task_name,
            "timestamp": timezone.now(),
        }
        if exception:
            args["exception"] = str(exception)
            if traceback := getattr(exception, "__traceback__"):
                args["traceback"] = "".join(
                    tb.format_exception(None, value=exception, tb=traceback)
                )
        return self.create(**args)


TaskLogManager = TaskLogManagerBase.from_queryset(TaskLogQuerySet)



# flake8: noqa
"""Script for adding lots of tasks to the queue for testing."""

import os
import sys
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

import django
from django.apps import apps

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

"""MAIN"""
from example.tasks import my_task

MAX_ENTRIES = 1_000

print(f"Adding {MAX_ENTRIES:,} tasks to the queue...")
for _ in range(MAX_ENTRIES):
    my_task.delay()
print("DONE!")



import json
from collections import defaultdict
from unittest import mock

import redis

from django.conf import settings
from django.test import TestCase

from taskmonitor.core import celery_queues

from ..factories import QueuedTaskRawFactory

CELERY_QUEUE_NAME = "test_task_monitor_celery"
MODULE_PATH = "taskmonitor.core.celery_queues"


def _redis_client():
    """Fetch the Redis client for the celery broker."""
    return redis.from_url(settings.BROKER_URL)


@mock.patch(MODULE_PATH + "._queue_base_name")
class TestCore(TestCase):
    def setUp(self):
        self.r = _redis_client()
        self._clear_queue()

    def tearDown(self):
        self._clear_queue()

    def _clear_queue(self):
        for queue_name in celery_queues._queue_names(CELERY_QUEUE_NAME):
            self.r.delete(queue_name)

    def _push_tasks(self, raw_tasks):
        tasks_by_priority = defaultdict(list)
        for task in raw_tasks:
            priority = task["properties"]["priority"]
            tasks_by_priority[priority].append(task)
        for priority, tasks in tasks_by_priority.items():
            raw_tasks_str = [json.dumps(obj) for obj in tasks]
            queue_name = f"{CELERY_QUEUE_NAME}{celery_queues.PRIORITY_SEP}{priority}"
            self.r.lpush(queue_name, *raw_tasks_str)

    def test_should_return_queue_length(self, mock_queue_base_name):
        # given
        mock_queue_base_name.return_value = CELERY_QUEUE_NAME
        raw_tasks = [QueuedTaskRawFactory(), QueuedTaskRawFactory()]
        self._push_tasks(raw_tasks)
        # when/then
        self.assertEqual(celery_queues.queue_length(), 2)

    def test_should_fetch_tasks_in_correct_order(self, mock_queue_base_name):
        # given
        mock_queue_base_name.return_value = CELERY_QUEUE_NAME
        raw_task_1 = QueuedTaskRawFactory(properties__priority=4)
        raw_task_2 = QueuedTaskRawFactory(properties__priority=4)
        raw_task_3 = QueuedTaskRawFactory(properties__priority=3)
        self._push_tasks([raw_task_1, raw_task_2, raw_task_3])
        # when
        result = celery_queues.fetch_tasks()
        # then
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], raw_task_3)
        self.assertEqual(result[1], raw_task_1)
        self.assertEqual(result[2], raw_task_2)



import itertools


class Echo:
    """An object that implements just the write method of the file-like
    interface.
    """

    def write(self, value):
        """Write the value by returning it, instead of storing in a buffer."""
        return value


def extract_app_name(task_name: str) -> str:
    """Extract the app name from a typical task name."""
    parts = task_name.split(".")
    try:
        idx = parts.index("tasks")
    except ValueError:
        if len(parts) == 2:
            return parts[0]
        else:
            return ""
    return parts[idx - 1] if idx > 0 else ""


def next_number(key: str = None) -> int:
    """Generate a sequence of numbers starting at 1.

    Args:
        key: key to generate sequence for.
    """
    if key is None:
        key = "_general"
    try:
        return next_number._counter[key].__next__()
    except AttributeError:
        next_number._counter = dict()
    except KeyError:
        pass
    next_number._counter[key] = itertools.count(start=1)
    return next_number._counter[key].__next__()
